#include <core.p4>
#define V1MODEL_VERSION 20200408
#include <v1model.p4>
#include "includes/defines.p4"
#include "includes/headers.p4"
#include "includes/parsers.p4"

//extern void prime();
//extern void prime2();

extern void packet_counter();

//extern void static_dispatch_function();

struct small_dig{
    bit<16> port;
}

struct digest_check_udp_port{
    bit<16> udp_port;
    bit<32> fid;
    bit<4> packet_count;
    bit<32> src_ip;
    bit<32> dst_ip;
    bit<8> autoscaling;
    bit<8> fno;
}

control ingress(inout headers hdr, inout metadata meta, inout standard_metadata_t standard_metadata) {

    //register<bit<8>>(10000) function_id_check; 
    register<bit<4>>(1) fwd_checks; 
    //bit<8> pc;
    bit<4> pc2=0;
    bit<1> static=1w1;

    @name(".fwd_act") action fwd_act(bit<16> port) {
        standard_metadata.egress_spec = port;
    }

    @name(".fwd") table fwd {
        actions = {
            fwd_act;
        }
        key = {
            standard_metadata.ingress_port : exact;
        }
    }

    @name(".dispatch_act") action dispatch_act(bit<32> dstAddr, bit<16> dstPort, bit<48> ethernetAddr, bit<16>egress_port, bit<8>autoscaling) {
        hdr.ipv4.dstAddr = dstAddr;
        hdr.udp.dstPort = dstPort;
        hdr.ethernet.dstAddr = ethernetAddr;
        hdr.map_hdr.autoscaling = autoscaling;
        //if(autoscaling>1)
        //{
        @atomic {
            packet_counter();
        }
        //}
    }
    #pragma netro no_lookup_caching dispatch_act

    @name(".dispatch") table dispatch {
        actions = {
            dispatch_act;
        }
        key = {
            hdr.map_hdr.function_id : exact;
        }
    }
    #pragma netro no_lookup_caching  dispatch

    @name(".warmstart_dispatch_act") action warmstart_dispatch_act(bit<32> dstAddr, bit<16> dstPort, bit<48> ethernetAddr , bit<16> egress_port) {
        hdr.ipv4.dstAddr = dstAddr;
        hdr.udp.dstPort = dstPort;
        hdr.ethernet.dstAddr = ethernetAddr;
    }
    #pragma netro no_lookup_caching warmstart_dispatch_act

     @name(".warmstart_dispatch") table warmstart_dispatch {
        actions = {
            warmstart_dispatch_act;
        }
        key = {
            hdr.map_hdr.function_id : exact;
            hdr.map_hdr.fno : exact;
        }
    }
    #pragma netro warmstart_dispatch no_lookup_caching 

    apply {
        if (hdr.ipv4.isValid() && hdr.udp.dstPort == DISPATCHER_PORT) {
            
            //function_id_check.read(pc,0);
            //pc = 8w2;
            //pc = hdr.map_hdr.function_id;
            //function_id_check.write(0,pc);
           
             @atmoic{

                if(hdr.map_hdr.fno==255) 
                {
                    hdr.map_hdr.function_id = 1;   
                }
                dispatch.apply();
                fwd_checks.read(pc2,0);
                pc2 = pc2 + 1;
                fwd_checks.write(0,pc2);

           
                if(hdr.map_hdr.autoscaling > 1)
                {      
                    warmstart_dispatch.apply();
                }
            }
            
            digest_check_udp_port dig;
            dig.udp_port = hdr.udp.dstPort;
            dig.fid = hdr.map_hdr.function_id;
            dig.packet_count = pc2;
            dig.src_ip = hdr.ipv4.srcAddr;
            dig.dst_ip = hdr.ipv4.dstAddr;
            dig.autoscaling = hdr.map_hdr.autoscaling;
            dig.fno = hdr.map_hdr.fno;

            digest<digest_check_udp_port>(0, dig );

            fwd.apply();

        } else {
            fwd.apply();
        }
        
        //bit<16>mod = 16w10;
        //hdr.udp.dstPort = 10000+(pc2 % mod);
        
        
    }
}

control egress(inout headers hdr, inout metadata meta, inout standard_metadata_t standard_metadata) {

    // @name(".ethernet_set_mac_act") action ethernet_set_mac_act(bit<48> smac, bit<48> dmac) {
    //     hdr.ethernet.srcAddr = smac;
    //     hdr.ethernet.dstAddr = dmac;
    // }
    // @name(".ethernet_set_mac") table ethernet_set_mac {
    //     actions = {
    //         ethernet_set_mac_act;
    //     }
    //     key = {
    //         standard_metadata.egress_port: exact;
    //     }
    // }

    @name("fix_checksum") action fix_checksum() {
        hdr.udp.checksum = 16w0;
    }
  

    apply {
        // if (hdr.udp.dstPort == MDS_PORT) {           
        //     ethernet_set_mac.apply();
        // }
        fix_checksum();
    }
}

control DeparserImpl(packet_out packet, in headers hdr) {
    apply {
        packet.emit<ethernet_t>(hdr.ethernet);
        packet.emit<ipv4_t>(hdr.ipv4);
        packet.emit<udp_t>(hdr.udp);
        packet.emit<map_hdr_t>(hdr.map_hdr);
    }
}

control verifyChecksum(inout headers hdr, inout metadata meta) {
    apply {
        verify_checksum(
            hdr.ipv4.isValid(),
            { hdr.ipv4.version,
            hdr.ipv4.ihl,
            hdr.ipv4.diffserv,
            hdr.ipv4.totalLen,
            hdr.ipv4.identification,
            hdr.ipv4.flags,
            hdr.ipv4.fragOffset,
            hdr.ipv4.ttl,
            hdr.ipv4.protocol,
            hdr.ipv4.srcAddr,
            hdr.ipv4.dstAddr },
            hdr.ipv4.hdrChecksum,
            HashAlgorithm.csum16);
    }
}

control computeChecksum(inout headers hdr, inout metadata meta) {
    apply {
        update_checksum(
            hdr.ipv4.isValid(),
            { hdr.ipv4.version,
            hdr.ipv4.ihl,
            hdr.ipv4.diffserv,
            hdr.ipv4.totalLen,
            hdr.ipv4.identification,
            hdr.ipv4.flags,
            hdr.ipv4.fragOffset,
            hdr.ipv4.ttl,
            hdr.ipv4.protocol,
            hdr.ipv4.srcAddr,
            hdr.ipv4.dstAddr },
            hdr.ipv4.hdrChecksum,
            HashAlgorithm.csum16);
    }
}

V1Switch<headers, metadata>(ParserImpl(), verifyChecksum(), ingress(), egress(), computeChecksum(), DeparserImpl()) main;
