let workerNodes = {}, timeline = {}
const constants = require('../constants_local.json')
const Heap = require('heap');
// const libSupport = require('./lib');
const dgram = require('dgram');
const udpProxy = dgram.createSocket('udp4');
// const logger = libSupport.logger;

const loadThreashold = 1
var workerHeap = new Heap(function(worker1, worker2) {
    return (worker1.system_info.avg_load[0] - worker2.system_info.avg_load[0]);
});

let coldstart_worker = {};

let resourceMap = {} // resource_id -> resource information map
let functionToResourceMap = {} // functionID -> resource_id map
let functionToResourceHeap = {} // functionID -> function Inastances heap map
let functionloadMap = {} // to store previous function instance having low load for functionID

let kafka = require('kafka-node'),
    Producer = kafka.Producer,
    client = new kafka.KafkaClient({
        kafkaHost: constants.network.external.kafka_host,
        autoConnect: true
    }),
    producer = new Producer(client),
    Consumer = kafka.Consumer,
    consumer = new Consumer(client,
        [
            { topic: constants.topics.heartbeat }, // receives heartbeat messages from workers, also acts as worker join message
            { topic: constants.topics.request_dm_2_rm }, // receives deployment details from RM
            { topic: constants.topics.check_autoscale },
            { topic: constants.topics.autoscale },
            { topic: constants.topics.deployed },
            { topic: constants.topics.coldstart_worker }, // give the information about worker having low load
            { topic: constants.topics.function_load},
            { topic: constants.topics.remove_worker},
            { topic: constants.topics.remove_function_intstance }
        ],
        [
            { autoCommit: true }
        ])

// console.log("consumer : ", consumer)

function getAddress() {
    return workerHeap.peek().address;
    // return Object.keys(workerNodes)[0];
}

// payloads = [
//     { topic: 'heartbeat', messages: 'hi', partition: 0 }
// ];
// producer.on('ready', function () {
//     producer.send(payloads, function (err, data) {
//         console.log(data);
//     });
// });

function updateColdstartOnNIC() {
    let lowloadworker = workerHeap.peek();
    if(lowloadworker.address != coldstart_worker.address)
    {
        console.log("min load worker changed !!!");
        coldstart_worker = lowloadworker;

        let payload = [{
            topic: constants.topics.coldstart_worker ,
            messages: JSON.stringify({
                    "resource_id": coldstart_worker.resource_id,
                    "timestamp": Date.now(),
                    "nodes": 
                        { 
                            node_id: coldstart_worker.address, 
                            portExternal: coldstart_worker.portExternal, 
                            mac: coldstart_worker.mac
                        }
            }),
            partition: 0
        }]
        producer.send(payload, () => {
            console.log(`\nupdate coldstart on NIC Replied`)
        })
    }
}


function packPacketFromDictionary(dataPacket) {

    let message = new Array(1024)
    let base = 0, chain_id, exec_id, function_id, data, function_count
    chain_id = struct.PackTo(">I", message, base, [dataPacket.chain_id])
    base += 4
    exec_id = struct.PackTo(">I", message, base, [dataPacket.exec_id])
    base += 4
    function_id = struct.PackTo(">I", message, base, [dataPacket.function_id])
    base += 4
    data = struct.PackTo(">I", message, base, [dataPacket.data])
    base += 4
    function_count = struct.PackTo("B", message, base, [dataPacket.function_count])
    base += 1
 
    // autoscale = struct.PackTo("B", message, base, [dataPacket.autostart])
    // base += 1
    // fno = struct.PackTo("B", message, base, [dataPacket.fno])
 
    f0 = struct.PackTo("B", message, base, [0])
    base += 1
    f1 = struct.PackTo("B", message, base, [12])
    base += 1
    f2 = struct.PackTo("B", message, base, [0])
    base += 1
    f3 = struct.PackTo("B", message, base, [34])
    base += 1
    f4 = struct.PackTo("B", message, base, [0])
    base += 1
 
  
    message = Buffer.from(message)
    return message
 }
 
 function compare(a, b) {
    return a.open_request_count - b.open_request_count
 }

 function autoscale_function(function_instance_info)
 {
     //TODO : Autoscale function from host DM
    //  i.e., send request DD from kafka to start new function
    // update entry of new instance in the f2r map and heap
    console.log("function instance info from autoscaling : ", function_instance_info)
    node_id = getAddress()
    let payload = [{
        topic: node_id,
        messages: JSON.stringify({
            "type": "execute", // Request sent to Dispatch Daemon via Kafka for actual deployment at the Worker
            resource_id: function_instance_info['resource_id'],
            runtime: function_instance_info['function_runtime'], 
            functionHash: function_instance_info['functionHash'],
            port: null, 
            resources: {
                memory: 0
            }
        }),
        partition: 0
    }]
    // logger.info(resourceMap);
    producer.send(payload, () => {
        console.log(`Autosacling Resource Deployment request sent to Dispatch Agent`)
    })
 }

 function downscaling_function(function_instance_info, resource_id)
 {
    console.log("downscaling function called !!!\n", function_instance_info)
    id = function_instance_info.functionHash + function_instance_info.function_runtime
    console.log("function closed, removing entry. ", function_instance_info.functionHash, resource_id)
    console.log("inside downscaling function , printing f2rap id : ", id, functionToResourceMap[id])
    functionToResourceMap[id][resource_id]['open_request_count'] = Number.MAX_SAFE_INTEGER
    functionToResourceHeap[id].updateItem(functionToResourceMap[id][resource_id])
    functionToResourceMap[id]['function_instance_count']--
    console.log("function map : ", functionToResourceMap[id])   
    // functionToResourceHeap[id] 
 }

 function updateNicRules()
 {
    // traverse functionloadmap hashmap 
    // send update message to smartnic monitor if function instance having low load changes
    ids = Object.keys(functionToResourceMap)
    console.log("\nupdating function rules on NIC \n\n")
    // console.log("\nfunction to REsource  MAP :  ",functionToResourceMap)
    // console.log("\nfunction to REsource  HEAP :  ",JSON.stringify(functionToResourceHeap))
    ids.forEach(id => {
        fmin = functionToResourceHeap[id].peek()

        // console.log("min finst : ", fmin)
        // console.log("loadmap : ", functionloadMap)
        if(fmin['resource_id'] != functionloadMap[id]['resource_id'])
        {
            let payload = [{
                topic: constants.topics.update_function_instance_nic,
                messages: JSON.stringify({
                        "resource_id": fmin.resource_id,
                        "functionHash": functionToResourceMap[id][fmin.resource_id]['functionHash'],
                        "portExternal": functionToResourceMap[id][fmin.resource_id]['portExternal'],
                        "mac": functionToResourceMap[id][fmin.resource_id]['mac'],
                        // "timestamp": Date.now(),
                        "node_id": fmin.node_id
                }),
                partition: 0
            }]
            producer.send(payload, () => {
                console.log(`\nsent update to smartNIC monitor\n`)
            })
            functionloadMap[id]= { 
                "node_id": fmin.node_id,
                "resource_id": fmin.resource_id
                // "load": 0,
                // "open_request_count": 0
            }
        }
        else{
            console.log("\nfunction instance with low load is not changed\n")
        }
    })
 }

 /**
 * function to check if any fuction is dead and stop sending heartbeat
**/
function check_function_heartbeat()
{
    console.log(" checking function heartbeat !!! function to resource map : ", functionToResourceMap)
    let ctime = Number(Date.now())
    Object.keys(functionToResourceMap).forEach( (fid) => {
    
        Object.keys(functionToResourceMap[fid]).forEach((rid) => {
            if(rid != "function_instance_count")
            {

                let data = functionToResourceMap[fid][rid]
                // console.log("INFORMATION : ", functionToResourceMap,functionToResourceMap[fid], data , data.timestamp)
                console.log("\n interval between heartbeat for function ", fid, " with  resource id : ",rid, " is : ", ctime - data.timestamp)
                if(ctime - data.timestamp > constants.heartbeat_threshold)
                {
                    // console.log("function ", fid," instance with rid ", rid, "is down", " data : ", data)
                    //function didn't sent heartbeat so close the remove function entry
                
                    let message = JSON.stringify({
                        functionHash: data.functionHash, portExternal: data.portExternal,
                        runtime: data.function_runtime, resource_id: rid, entity_id: 0,
                        total_request: 0, wait_time: 0
                    })
                    // console.log("\nmessage to remove function entry from all components  : ", message)
            
                    console.log("\nmessage to remove function entry from all components ")
                    
                    // console.log("Not received heartbeat for too long. Exiting");
                    if (producer)
                        producer.send(
                            [
                                {
                                    topic: constants.topics.remove_function_intstance, 
                                    messages: message 
                                }
                            ], (err, data) => {
                                if (err)
                                    console.log(err);
                                    // 
                                console.log("Removing entry for function", data.functionHash, "resource_id", data.resource_id);
                            
                            })
                }
            }
        })
    })
} 

consumer.on('message', function (message) {

    let topic = message.topic
    message = message.value
    // console.log("message ",message)
    // console.log("function map : ", functionToResourceMap)
    // console.log("function heap : ", functionToResourceHeap)
    
    // if (topic !== "heartbeat2")
    //     console.log("not a heartbeat message : ", message);
    if (topic === constants.topics.heartbeat) {
        message = JSON.parse(message)
        if (Date.now() - message.timestamp < 1000)
        {
            if (!workerNodes[message.address]) {
                workerNodes[message.address] = message
                console.log("New worker discovered. Worker List: ")
                console.log(workerNodes);
                workerHeap.push(workerNodes[message.address])
                if(Object.keys(workerNodes).length === 1)
                {
                    updateColdstartOnNIC();
                }
            }
            else{
                // console.log("Got heartbeat updating load of wroker ", message.address)
                if(Math.abs(workerNodes[message.address].system_info.avg_load[0] - message.system_info.avg_load[0]) > loadThreashold )
                {
                    workerNodes[message.address].system_info = message.system_info
                    workerNodes[message.address].timestamp = message.timestamp
                    console.log("updated wroker load : ", workerNodes[message.address])
                    workerHeap.updateItem(workerNodes[message.address])
                    updateColdstartOnNIC();
                }
                else
                {
                    // console.log("Change in worker load is less than threshold")
                    //f
                }
            }
            // console.log("\nheap : ",workerHeap)
        }

    } else if (topic === constants.topics.request_dm_2_rm) {
        message = JSON.parse(message)
        console.log(message);
        
        let payload = [{
            topic: "RESPONSE_RM_2_DM_DUMMY2",
            messages: JSON.stringify({
                    "resource_id": message.resource_id,
                    "timestamp": Date.now(),
                    "nodes": [
                        { node_id: getAddress(), port: null }]
            }),
            partition: 0
        }]
        producer.send(payload, () => {
            console.log(`Replied`)
        })
    } else if (topic === constants.topics.check_autoscale) {
        message = JSON.parse(message)
        console.log("request to check autoscale : ", message);

    } else if ( topic == constants.topics.autoscale) {
        message = JSON.parse(message)
        console.log("request to scale : ", message);
        let request_id = Math.floor(Math.random() * 1000)
        // req.body.request_id = request_id
        // res.request_id = request_id
        // requestFlightQueue.set(request_id, res)
        // let payload = "HELLO"
        // payload.request_id = request_id
        // let data = payload.data
        let fid= message.functionHash.slice(9) 
        // res.data_set_time = Date.now()
        let packet = packPacketFromDictionary({
            chain_id: 0,
            exec_id: request_id,
            function_id: fid,
            data,
            function_count: 1,
            // autostart: 1,
            // fno : 255
        })
        // res.pack_time = Date.now()
        // udpProxy.send(packet, 0, packet.length, "8000", "192.168.2.2", function (err, bytes) {
        //     logger.info(`forwarded request via UDP, IP 192.168.2.2 Port 8000`)
        //     res.send_time = Date.now()
        // })
        // udpProxy.send(packet, 0, packet.length, resource.port, resource.node_id, function (err, bytes) {
        //     logger.info(`forwarded request via UDP, IP 192.168.2.5 Port ${resource.port}`)
        //     res.send_time = Date.now()
        // })

    } else if (topic == constants.topics.deployed) {
        // deployed request will be handled to deploy new function insstance 
        message = JSON.parse(message)
        //fuction is deployed successfully
        // if function map availabe add it in the heap
        // if not availabe create new heap map
        console.log("message deployed : ", message)
        let id = message.functionHash + message.runtime
        // if (resourceMap.has(message.resource_id)) {
        //     let resource = resourceMap.get(message.resource_id)
        //     resource.node_id = message.node_id.trim()
        // }
        resource_id = message.resource_id

        resourceMap[message.functionHash + resource_id]= message
        if (functionToResourceMap[id]) {
            console.log("\nfunction ID already has a function heap so inserting into heap!!!\n")
            // let functionResources = functionToResourceMap[id]
            // let functionResourcesHeap = functionToResourceHeap[id]
            // functionResources.push({
            //     resource_id: message.resource_id
            // })
            // functionToResourceMap[id]={}
            let ctime = Date.now()
            console.log("\n\nctime : ", ctime)
            functionToResourceMap[id][resource_id] = {
                "functionHash": message.functionHash,
                "resource_id": resource_id,
                "function_runtime": message.runtime,
                "node_id": message.node_id, 
                "portExternal": message.portExternal,
                "mac": message.mac,
                "open_request_count": 0,
                "load": 0 ,
                "timestamp" : Number(ctime)
            }
            functionToResourceMap[id].function_instance_count++
            functionToResourceHeap[id].push(functionToResourceMap[id][resource_id])
            // heap.push(functionResources, {
            //     resource_id: message.resource_id,
            //     open_request_count: 0,
            //     cpu_utilization: 0
            // }, libSupport.compare_uti)
            console.log("Horizontally scaling up: " +JSON.stringify(functionToResourceHeap[id]));
            console.log("function to resource MAP : ", functionToResourceMap)
    
        } else {
            
            console.log("\nthere is no function heap for function ID so creating one\n")
            // resourceMap[message.functionHash + message.resource_id] = {
            //     "resource_id" : message.resource_id, 
            //     "function_runtime": message.runtime,
            //     "node_id": message.node_id
            // }
            functionToResourceMap[id] = {
                function_instance_count: 1
            }
            // functionToResourceMap[id][function_instance_count]=1
            let ctime = Date.now()
            console.log("\n\nctime : ", ctime)
            functionToResourceMap[id][resource_id] = {
                "functionHash": message.functionHash,
                "resource_id": resource_id,
                "function_runtime": message.runtime,
                "node_id": message.node_id, 
                "portExternal": message.portExternal,
                "mac": message.mac,
                "open_request_count": 0,
                "load": 0 ,
                "timestamp": Number(ctime)    
            }
            
            functionloadMap[id]= { 
                "functionHash": message.functionHash,
                "node_id": message.node_id,
                "resource_id": resource_id,
                "load": 0,
                "open_request_count": 0,
                "high_load_count": 0,
                "low_load_count": 0
            }
            functionToResourceHeap[id] = new Heap(compare)
            functionToResourceHeap[id].push(functionToResourceMap[id][resource_id])
            // functionToResourceMap.set(id, resourceHeap)
            // functionResources = heap(compare)
            // functionToResourceMap[id]= functionResources
            // console.log("Creating new resource pool"+ JSON.stringify(functionToResourceHeap[id]));
            console.log("Creating new resource pool"+ functionToResourceHeap[id]);
            console.log("function to resource MAP : ", functionToResourceMap)
            console.log('function instance count : ',functionToResourceMap[id]['function_instance_count'])
        }
    }
    else if(topic == constants.topics.function_load)
    {
        message = JSON.parse(message)
        resource_id = message.resource_id
        id = message.functionHash + message.runtime
        if(functionToResourceHeap[id])
        {
            // console.log("functionToResourceHeap id is present updating heap")
            functionToResourceMap[id][resource_id]["timestamp"] = Number(Date.now())
            functionToResourceMap[id][resource_id]['open_request_count'] = message.open_request_count
            functionToResourceHeap[id].updateItem(functionToResourceMap[id][resource_id])
            if(functionToResourceMap[id][resource_id]['open_request_count'] >= constants.autoscalar_metrics.high_open_request_threshold){
                // autoscale_function(functionToResourceMap[id][message.resource_id])
                // console.log
                functionloadMap[id]['high_load_count']++
                if(functionloadMap[id]['high_load_count'] >= constants.autoscalar_metrics.high_load_count)
                {
                    console.log("\n\nfunction load increases. UP scale the function\n\n")
                    //autoscale function : deploy new function instance for this function ID
                    functionloadMap[id]['high_load_count']=0
                    autoscale_function(functionToResourceMap[id][resource_id])
                }
            }
            else{
                functionloadMap[id]['high_load_count']=0
            }
            if(functionToResourceMap[id][resource_id]['open_request_count'] <= constants.autoscalar_metrics.low_open_request_threshold
                && functionToResourceMap[id].function_instance_count>1){
                
                functionloadMap[id]['low_load_count']++
                if(functionloadMap[id]['low_load_count'] >= constants.autoscalar_metrics.low_load_count)
                {
                    //downscale function: delete any function instance for this function ID if number of function instance > 1
                
                    console.log("\n\nfunction load decreases. DOWN scale the function\n\n")
                    functionloadMap[id]['low_load_count']=0
                    downscaling_function(functionToResourceMap[id][resource_id], resource_id)

                }
            }else{
                functionloadMap[id]['low_load_count']=0
            }
        }
        else{
            console.log("functionToResourceHeap id is NOT present NOT updating heap")
        }

        // console.log("\n\nupdated function heap ", functionToResourceHeap)
    }
    else if(topic == constants.topics.remove_function_intstance)
    {
        message = JSON.parse(message)
        console.log("remove function instance called !!!\n", message)
        id = message.functionHash + message.runtime
        console.log("function closed, removing entry. ", message.functionHash, message.resource_id)
        if(functionToResourceMap[id])
        {
            functionToResourceMap[id][message.resource_id]['open_request_count'] = Number.MAX_SAFE_INTEGER
            functionToResourceHeap[id].updateItem(functionToResourceMap[id][message.resource_id])
            functionToResourceMap[id]['function_instance_count']--
            if(functionToResourceHeap[id].peek()['open_request_count']==Number.MAX_SAFE_INTEGER )
            {
                console.log("all function instance of this functionid are stopped. so removing entry from map and heap")
                delete functionToResourceHeap[id]
                delete functionToResourceMap[id]
                rmv_worker_message = JSON.stringify(message)

                producer.send(
                    [
                        {topic: "removeWorker2", messages: rmv_worker_message }
                    ], (err, data) => {
                        if (err)
                            console.log(err);
                            
                        console.log("Ending worker for function", message.functionHash, "resource_id", message.resource_id);

                    })
            }
        }else{
              console.log("function map for that id is not present!!")
        }
        console.log("rfi function map : ", functionToResourceMap)   
        // functionToResourceHeap[id]
    }
    // else if(topic == constants.topics.remove_worker)
    // {
    //     message = JSON.parse(message)
    //     console.log("remove function called !!!\n", message)
    //     id = message.functionHash + message.runtime
    //     console.log("function closed, removeing entry. ", message.functionHash, message.resource_id)
    //     functionToResourceMap[id][message.resource_id]['open_request_count'] = Number.MAX_SAFE_INTEGER
    //     functionToResourceHeap[id].updateItem(functionToResourceMap[id][message.resource_id])
    //     functionToResourceMap[id]['function_instance_count']--
    //     if(functionToResourceHeap[id].peek()['open_request_count']==Number.MAX_SAFE_INTEGER )
    //     {
    //         console.log("all function instance of this functionid are stopped. so removing entry from map and heap")
    //         delete functionToResourceHeap[id]
    //         delete functionToResourceMap[id]
    //     }
    //     console.log("function map : ", functionToResourceMap[id])   
    //     // functionToResourceHeap[id]
    // }
})
setInterval(updateNicRules, 5000)
setInterval(check_function_heartbeat, 5000)
